import copy
from typing import List, Dict

from hpbandster.optimizers.config_generators.bohb import BOHB

from hpogrid.generator.base_generator import Generator
from hpogrid.search_space.bohb_space import BOHBSpace

default_budget = 100

class BOHBResultWrapper():
    def __init__(self, config, loss, budget=default_budget):
        self.result = {"loss": loss}
        self.kwargs = {"budget": budget, "config": config.copy()}
        self.exception = None

class BOHBGenerator(Generator):

    def get_searcher(self, search_space:Dict, metric:str, mode:str, **args):
        search_space = BOHBSpace(search_space).get_search_space()
        searcher = BOHB(search_space)
        return searcher

    def ask(self, n_points:int = None):
        points = []
        for _ in range(n_points):
            point, info = self.searcher.get_config(None)
            points.append(copy.deepcopy(point))
        return points

    def tell(self, point:Dict, value):
        value = self._to_metric_values(value)
        result = BOHBResultWrapper(point, self.signature * value)
        self.searcher.new_result(result)






from skopt import Optimizer

import ray
from ray.tune.suggest.skopt import SkOptSearch

from hpogrid.search_space.skopt_space import SkOptSpace



class SkOptAlgoWrapper():
	def __init__(self):
		self.algorithm = None
	def create(self, metric, mode, search_space, **args):
		search_space = SkOptSpace(search_space).search_space
		optimizer = Optimizer(search_space)
		labels = [hp.name for hp in search_space]

		self.algorithm = SkOptSearch(optimizer, labels, metric=metric, mode=mode, **args)
		return self.algorithm


import json
import os

try:
    from hpogrid.components.defaults import *
except:
    raise ImportError('ERROR: Cannot import hpogrid module. Try source setupenv first.')



kHPSamplingFormat = {
	'categorical': {
		'categories' : { 'type': list, 'required': True},
		'grid_search' : { 'type': int, 'required': False, 'default': 0}
	},
	'uniform': {
		'low': { 'type': (float, int), 'required': True},
		'high': { 'type': (float, int), 'required': True}
	},
	'uniformint': {
		'low': { 'type': int, 'required': True},
		'high': { 'type': int, 'required': True}
	},
	'quniform': {
		'low': { 'type': (float, int), 'required': True},
		'high': { 'type': (float, int), 'required': True},
		'q': { 'type': (float, int), 'required': True}
	},
	'loguniform': {
		'low': { 'type': (float, int), 'required': True},
		'high': { 'type': (float, int), 'required': True},
		'base': { 'type': (float, int), 'required': False}
	},
	'qloguniform': {
		'low': { 'type': (float, int), 'required': True},
		'high': { 'type': (float, int), 'required': True},
		'q': { 'type': (float, int), 'required': True},
		'base': { 'type': (float, int), 'required': False}
	},
	'normal': {
		'mu': { 'type': (float, int), 'required': True},
		'sigma': { 'type': (float, int), 'required': True}
	},
	'qnormal': {
		'mu': { 'type': (float, int), 'required': True},
		'sigma': { 'type': (float, int), 'required': True},
		'q': { 'type': (float, int), 'required': True}
	},
	'lognormal': {
		'mu': { 'type': (float, int), 'required': True},
		'sigma': { 'type': (float, int), 'required': True},
		'base': { 'type': (float, int), 'required': False}
	},
	'qlognormal': {
		'mu': { 'type': (float, int), 'required': True},
		'sigma': { 'type': (float, int), 'required': True},
		'base': { 'type': (float, int), 'required': False},
		'q': { 'type': (float, int), 'required': True}
	},
	'fixed': {
		'value': { 'type': None, 'required': True}
	}
}

kGridConfigFormat = {
	'site': {
		'required': True,
		'type': str,
		'choice': kGPUGridSiteList,
		'default': kDefaultGridSite
	},
	'container': {
		'required': True,
		'type': str,
		'default': kDefaultContainer
	},
	'retry': {
		'required': False,
		'type': int,
		'default': 0
	},
	'inDS': {
		'required': True,
		'type': str
	},
	'outDS': {
		'required': True,
		'type': str,
		'default': kDefaultOutDS
	}

}

kHPOConfigFormat = {
	'algorithm': {
		'required': True,
		'type': str,
		'choice': kSearchAlgorithms
	},
	'metric' : {
		'required': True,
		'type': str,
	},
	'mode' : {
		'required': True,
		'type': str,
		'choice': kMetricMode,
	},
	'scheduler': {
		'required': False,
		'type': str,
		'choice': kSchedulers,
		'default': kDefaultScheduler
	},
	'trials': {
		'required': False,
		'type': int,
		'default': 100
	},
	'log_dir': {
		'required': False,
		'type': str,
		'default': './log'
	},
	'verbose': {
		'required': False,
		'type': int,
		'default': 0
	}
}

def type2str(typevar):
	if isinstance(typevar, tuple):
		return ', '.join([t.__name__ for t in typevar])
	else:
		return typevar.__name__

def validate_project_env_path():
	if kProjectPath in os.environ:
		return True
	else:
		print('ERROR: Path variable {} not set.'
			'Maybe you need to source setupenv.sh first.'.format(kProjectPath))
		return False


def validate_hpogrid_env_path():
	if kHPOGridEnvPath in os.environ:
		return True
	else:
		print('ERROR: Path variable {} not set.'
			'Maybe you need to source setupenv.sh first.'.format(kHPOGridEnvPath))
		return False

def validate_hpo_config(config):
	print('Info: Validating HPO configuration...')
	config_format = kHPOConfigFormat
	for key in config_format:
		if key in config:
			# check if the value type of the config is correct
			value_type = config_format[key]['type']
			if not isinstance(config[key], value_type):
				print('ERROR: The value of {} must be of type {}'.format(key, type2str(value_type)))
				return False
			# check if the value of the config is supported
			if ('choice' in config_format[key]) and (config[key] not in config_format[key]['choice']):
				print('ERROR: The value of {} must be one of the followings')
				print(str(config_format[key]['choice']).strip('[]'))
				return False
		else:
			if config_format[key]['required']:
				print('ERROR: The required key {} is missing from the configuration'.format(key))
				return False
			# fill in default config if not specified
			if 'default' in config_format[key]:
				print('Info: Added the key {} with default value {} to the configuration '.format(
					key, str(config_format[key]['default'])))
				config[key] = config_format[key]['default']

	print('Info: Successfully validated HPO configuration')
	return True

def validate_search_space(search_space):
	print('Info: Validating search space configuration...')
	sampling_methods = kHPSamplingFormat.keys()
	kHPKeys = set(['method', 'dimension'])
	for hp in search_space:
		if (not isinstance(search_space[hp], dict)) or (set(search_space[hp].keys()) != kHPKeys):
			print('ERROR: Each hyperparameter must be defined by a dictionary containing the keys "method" and "dimension"')
			return False	
		if not isinstance(search_space[hp]['dimension'], dict):
			print('ERROR: Dimension of a hyperparameter must be a dictionary containing the parametrs of its sampling method')
			return False
		method = search_space[hp]['method']
		# check if hyperparameter sampling method is supported
		if method not in sampling_methods:
			print('ERROR: Sampling method {} is not supported'.format(method))
			return False
		for arg in kHPSamplingFormat[method]:
			if arg in search_space[hp]['dimension']:
				# check if the argument type of the sampling method is correct
				value_type = kHPSamplingFormat[method][arg]['type']
				if not isinstance(search_space[hp]['dimension'][arg], value_type):
					print('ERROR: The value of argument {} for the sampling method {} must be of type {}'.format(
						arg, method, type2str(value_type)))
					return False
			else:
				# check if a required argument for the sampling method is missing
				if (kHPSamplingFormat[method][arg]['required']):
					print('ERROR: Missing argument {} for the sampling method {}'.format(arg, method))
					return False
				# set the argument to its default value if not specified
				if ('default' in kHPSamplingFormat[method][arg]):
					print('Info: The argument {} for the sampling method {} will be set to its default value {}'.format(
						arg, method, kHPSamplingFormat[method][arg]['default']))
					search_space[hp]['dimension'][arg] = kHPSamplingFormat[method][arg]['default']
	print('Info: Successfully validated search space configuration')
	return True



def validate_grid_config(config):
	return True


def validate_job_metadata(data):
	if not isinstance(data, dict):
	    return False
	return all(key in kHPOGridMetadataFormat for key in data)


import os
import sys
import json
from typing import List, Dict

import argparse

kDefaultGenerator = 'nevergrad'
kDefaultMetric = 'loss'
kDefaultMode = 'min'

class NpEncoder(json.JSONEncoder):
    def default(self, obj):
        if isinstance(obj, np.integer):
            return int(obj)
        elif isinstance(obj, np.floating):
            return float(obj)
        elif isinstance(obj, np.ndarray):
            return obj.tolist()
        else:
            return super(NpEncoder, self).default(obj)

class SteeringIDDS():
    def __init__(self, parse=True):
        if (len(sys.argv) > 1) and parse:
            self.run_parser()

    def get_parser(self):
        parser = argparse.ArgumentParser(
                    formatter_class=argparse.RawDescriptionHelpFormatter)
        parser.add_argument('-l', '--lib', default=kDefaultGenerator,
            help='library to use')  
        parser.add_argument('-s', '--space',
            help='search space json file')
        parser.add_argument('-n', '--n_point', type=int, required=True,
            help='number of points to generate')
        parser.add_argument('-m', '--max_point', type=int,
            help='maximum number of points to generate in entire iDDS workflow')
        parser.add_argument('-e', '--metric', default=kDefaultMetric,
            help='evaluation metric')
        parser.add_argument('-d', '--mode', default=kDefaultMode,
            help='evaluation mode')
        parser.add_argument('-i', '--infile', help='iDDS input')
        parser.add_argument('-o', '--outfile', help='iDDS output')
        parser.add_argument('--method', help='optimizer type (nevergrad only)')
        return parser  

    def get_generator(self, space, metric=kDefaultMetric, mode=kDefaultMode, lib=kDefaultGenerator, **args):
        if lib == 'nevergrad':
            from hpogrid.generator.nevergrad_generator import NeverGradGenerator
            return NeverGradGenerator(space, metric, mode, **args)
        elif lib == 'skopt':
            from hpogrid.generator.skopt_generator import SkOptGenerator
            return SkOptGenerator(space, metric, mode, **args)
        elif lib == 'hyperopt':
            from hpogrid.generator.hyperopt_generator import HyperOptGenerator
            return HyperOptGenerator(space, metric, mode, **args)
        elif lib == 'ax':
            from hpogrid.generator.ax_generator import AxGenerator
            return AxGenerator(space, metric, mode, **args)
        elif lib == 'bohb':
            from hpogrid.generator.bohb_generator import BOHBGenerator
            return BOHBGenerator(space, metric, mode, **args) 
        else:
            raise ValueError('Generator from library {} is not supported'.format(lib))

    def run_parser(self):
        parser = self.get_parser()
        if os.path.basename(sys.argv[0]) == 'hpogrid':
            args = parser.parse_args(sys.argv[2:])
        else:
            args = parser.parse_args(sys.argv[1:])
        self.generate_points(**vars(args))

    def validate_idds_input(self, idds_input):
        keys = ['points']
        for key in keys:
            if key not in idds_input:
                raise KeyError('Key {} not found in idds input.'
                    ' Please check idds input format'.format(key))
        if not isinstance(idds_input['points'], List):
            raise ValueError('idds should have the data structure: Dict[List[...]]')

        for point in idds_input['points']:
            if not isinstance(point, List):
                raise ValueError('idds should have the data structure: Dict[List[List[...]]]')
            elif not isinstance(point[0], Dict):
                raise ValueError('idds should have the data structure: Dict[List[List[Dict, Value]]]')
        return True

    def parse_idds_input(self, file):
        if not file:
            return None, None, None
        points = []
        results = []
        search_space = None

        with open(file,'r') as input_file:
            idds_data = json.load(input_file)
        self.validate_idds_input(idds_data)
        
        input_data = idds_data['points']
        for data in input_data:
            if data[1]:
                points.append(data[0])
                results.append(data[1])


        search_space = None if 'opt_space' not in idds_data else idds_data['opt_space']

        return points, results, search_space

    def generate_points(self, space, n_point, metric=kDefaultMetric, mode=kDefaultMode,
        lib=kDefaultGenerator, max_point=None, infile=None, outfile=None, **args):
        points, results, idds_search_space = self.parse_idds_input(infile)

        # load search space
        if (not space) and (not idds_search_space):
            raise RuntimeError('search space is not specified in either idds input or '
            'through command line argument')
        if idds_search_space:
            search_space = idds_search_space            
        if space:
            with open(space, 'r') as space_input:
                search_space = json.load(space_input)

        # create generator
        generator = self.get_generator(search_space, metric, mode, lib, **args)

        n_evaluated = 0

        # feed evaluated points
        if points and results:
            generator.feed(points=points, results=results)
            print('INFO: Feeded the following results to the {} optimizer'.format(lib))
            generator.show(points=points, results=results)
            n_evaluated = len(points)

        # determine number of points to generate
        max_point = 9999 or max_point
        n_remaining = max_point-n_evaluated
        if n_remaining < 0:
            raise ValueError('there are already more evaluated points than the'
            ' maximum points to generate for idds workflow')
        n_generate = min(n_point, n_remaining)

        # generate points
        new_points = generator.ask(n_generate)
        print('INFO: Generated {} new points'.format(n_generate))
        generator.show(new_points)

        # save output
        if outfile:
            with open(outfile, 'w') as out:
                json.dump(new_points, out, indent=2, cls=NpEncoder)





import os
import sys
import json
import time
import importlib
import argparse
from datetime import datetime
from pdb import set_trace

import ray
from ray import tune

try:
    from hpogrid.components import validation
    from hpogrid.components.defaults import *
    from hpogrid.utils import helper
except:
    raise ImportError('Cannot import hpogrid module. Try source setupenv first.')
    

class JobHandler():
    def __init__(self, parse=True):
        # run hyperparameter optimization for a project via hpogrid executable
        if (len(sys.argv) > 1) and parse:
            self.run_parser()

    def get_parser(self):
        parser = argparse.ArgumentParser(
                    formatter_class=argparse.RawDescriptionHelpFormatter)
        parser.add_argument('proj_name', help='the project to run hyperparameter optimization')
        return parser    

    def run_parser(self):
        parser = self.get_parser()
        if os.path.basename(sys.argv[0]) == 'hpogrid':
            args = parser.parse_args(sys.argv[2:])
        else:
            args = parser.parse_args(sys.argv[1:])
        self.run_project(args.proj_name)

    def get_scheduler(self, name, metric, mode, search_space = None, **args):
        if (name == None) or (name == 'None'):
            return None
        elif name == 'asynchyperband':
            from hpogrid.scheduler.asynchyperband_scheduler import AsyncHyperBandSchedulerWrapper
            return AsyncHyperBandSchedulerWrapper().create(metric, mode, **args)
        elif name == 'bohbhyperband':
            from hpogrid.scheduler.bohbhyperband_scheduler import BOHBHyperBandSchedulerWrapper
            return BOHBHyperBandSchedulerWrapper().create(metric, mode, **args)
        elif name == 'pbt':
            from hpogrid.scheduler.pbt_scheduler import PBTSchedulerWrapper
            if search_space is None:
                raise ValueError('search space must be set before using pbt scheduler')
            return PBTSchedulerWrapper().create(metric, mode, search_space, **args)

    def rename_algorithm(self, algo):
        if algo == 'random':
            algo = 'tune'
        elif algo == 'bayesian':
            algo = 'skopt'
        return algo
        
    def get_search_space(self, raw_space, algo):
        if raw_space is None:
            raise ValueError('search space can not be empty')    
        algo = self.rename_algorithm(algo)
        if algo == 'ax':
            from hpogrid.search_space.ax_space import AxSpace
            return AxSpace(raw_space).get_search_space()
        elif algo == 'bohb':
            from hpogrid.search_space.bohb_space import BOHBSpace
            return BOHBSpace(raw_space).get_search_space()
        elif algo == 'hyperopt':
            from hpogrid.search_space.hyperopt_space import HyperOptSpace
            return HyperOptSpace(raw_space).get_search_space()
        elif algo == 'skopt':
            from hpogrid.search_space.skopt_space import SkOptSpace
            return SkOptSpace(raw_space).get_search_space()
        elif algo == 'tune':
            from hpogrid.search_space.tune_space import TuneSpace
            return TuneSpace(raw_space).get_search_space()
        elif algo == 'nevergrad':
            from hpogrid.search_space.nevergrad_space import NeverGradSpace
            return NeverGradSpace(raw_space).get_search_space() 
        else:
            raise ValueError('the algorithm {} is not supported'.format(algo))

    def get_algorithm(self, algo, metric, mode, space, **args):
        algo = self.rename_algorithm(algo)
        if algo == 'ax':
            from hpogrid.algorithm.ax_algorithm import AxAlgoWrapper
            return AxAlgoWrapper().create(metric, mode, space, **args)
        elif algo == 'bohb':
            from hpogrid.algorithm.bohb_algorithm import BOHBAlgoWrapper
            return BOHBAlgoWrapper().create(metric, mode, space, **args)
        elif algo == 'hyperopt':
            from hpogrid.algorithm.hyperopt_algorithm import HyperOptAlgoWrapper
            return HyperOptAlgoWrapper().create(metric, mode, space, **args)
        elif algo == 'skopt':
            from hpogrid.algorithm.skopt_algorithm import SkOptAlgoWrapper
            return SkOptAlgoWrapper().create(metric, mode, space, **args)
        elif algo == 'nevergrad':
            from hpogrid.algorithm.nevergrad_algorithm import NeverGradAlgoWrapper
            return NeverGradAlgoWrapper().create(metric, mode, space, **args)
        elif algo == 'tune':
            return None
        else:
            raise ValueError('the algorithm {} is not supported'.format(algo))

    def get_model(self, script_name, model_name):
        model = None
        script_name_noext = os.path.splitext(script_name)[0]
        try: 
            module = importlib.import_module(script_name_noext)
            model = getattr(module, model_name)
        except: 
            raise ImportError('Unable to import function/class {} '
                'from training script: {}.py'.format(model_name, script_name_noext))
        return model

    def save_metadata(self, df, summary):
        hyperparameters = summary['hyperparameters']
        rename_hp = { 'config/{}'.format(hp): hp for hp in hyperparameters}
        df = df.rename(columns=rename_hp)
        df = df.rename(columns={'time_total_s':'time_s'})
        columns_to_save = ['time_s'] + hyperparameters
        if 'metric' in summary:
            columns_to_save.append(summary['metric'])
        df = df.filter(columns_to_save, axis=1).transpose()
        result = df.to_dict()
        summary['result'] = result

        with open(kGridSiteMetadataFileName,'w') as output:
            json.dump(summary, output, cls=helper.NpEncoder)

    @staticmethod
    def get_resource_info():
        resource = {}
        n_gpu = helper.get_n_gpu()
        n_cpu = helper.get_n_cpu()
        print('INFO: Number of GPUs detected: ',n_gpu)
        print('INFO: Number of CPUs detected: ',n_cpu)
        if n_gpu > 0:
            resource['gpu'] = n_gpu
            print('INFO: Each trial will use {} GPU(s) resource'.format(n_gpu))
        if n_cpu > 0:
            resource['cpu'] = n_cpu
            print('INFO: Each trial will use {} CPU(s) resource'.format(n_cpu))
        if not resource:
            resource = None
        return resource

    def run(self, hpo_config, search_space, model_config, proj_name=None):
        start = time.time()

        algorithm = self.rename_algorithm(hpo_config['algorithm'])

        if algorithm == 'tune':
            tune_config_space = self.get_search_space(search_space, algo='tune')
        else:
            tune_config_space = {}

        tune_config_space.update(model_config['param'])

        algorithm = self.get_algorithm(
            algorithm,
            hpo_config['metric'],
            hpo_config['mode'],
            search_space, **hpo_config['algorithm_param'])
    
        scheduler = self.get_scheduler(
            hpo_config['scheduler'],
            hpo_config['metric'],
            hpo_config['mode'],
            search_space, **hpo_config['scheduler_param'])

        resources_per_trial = JobHandler.get_resource_info()

        model = self.get_model( model_config['script'], model_config['model'])

        #ray.init(node_ip_address="127.0.0.1", ignore_reinit_error=True)
        ray.init(ignore_reinit_error=True)

        start_data_time = datetime.now().strftime("%m/%d/%Y, %H:%M:%S")

        analysis = tune.run(
            model,
            name=proj_name,
            scheduler=scheduler,
            search_alg=algorithm,
            config=tune_config_space,
            num_samples=hpo_config['trials'],
            resources_per_trial=resources_per_trial,
            verbose=hpo_config['verbose'],
            local_dir=hpo_config['log_dir'],
            stop=hpo_config['stop'],
            raise_on_failed_trial=False)

        end = time.time()
        total_time = float(end-start)
        end_data_time = datetime.now().strftime("%m/%d/%Y, %H:%M:%S")

        best_config = analysis.get_best_config(metric=hpo_config['metric'], mode = hpo_config['mode'])
        print("Best config: ", best_config)
        print("Time taken in seconds: ", total_time)

        df = analysis.dataframe()

        summary = {
            'title' : proj_name,
            'start_date_time': start_data_time,
            'end_date_time': end_data_time,
            'task_time_s' : total_time,
            'hyperparameters': list(search_space.keys()),
            'metric': hpo_config['metric'],
            'mode' : hpo_config['mode'],
            'best_config' : best_config, 
        }

        self.save_metadata(df, summary)
    
        ray.shutdown()

    def run_project(self, proj_name):
        
        hpo_config = helper.get_hpo_config(proj_name)

        model_config = helper.get_model_config(proj_name)

        search_space = helper.get_search_space_config(proj_name)

        helper.set_script_paths(proj_name)

        self.run(hpo_config, search_space, model_config, proj_name)

        helper.set_script_paths(proj_name, undo=True)



if __name__ == '__main__':
    parser = argparse.ArgumentParser()
    parser.add_argument('proj_name')
    args = parser.parse_args()
    JobHandler().run_project(args.proj_name)



import copy
from typing import List, Dict

from ax.service.ax_client import AxClient

from hpogrid.generator.base_generator import Generator
from hpogrid.search_space.ax_space import AxSpace

class AxGenerator(Generator):
    def get_searcher(self, search_space:Dict, metric:str, mode:str):
        ax_client = AxClient(enforce_sequential_optimization=False, verbose_logging=False)
        search_space = AxSpace(search_space).get_search_space()
        if mode == 'max':
            minimize = False
        elif mode == 'min':
            minimize = True
        else:
            raise ValueError('mode of evaluation metric can only be "min" or "max"')

        ax_client.create_experiment(
            parameters=search_space,
            objective_name=metric,
            minimize=minimize)
        return ax_client

    def tell(self, point:Dict, value):
        value = self._to_metric_values(value)
        _, trial_index = self.searcher.attach_trial(point)
        metric_dict = { self.metric: (value, 0.0)}
        self.searcher.complete_trial(
            trial_index=trial_index, raw_data=metric_dict)

    def ask(self, n_points:int =None):
        points = []
        for _ in range(n_points):
            point, trial_index = self.searcher.get_next_trial()
            trial = self.searcher._get_trial(trial_index=trial_index)
            trial.mark_abandoned()
            points.append(copy.deepcopy(point))
        return points


import sys, os
import argparse
import json
import shutil
#import re
import fnmatch


import pandas as pd 
from pdb import set_trace
from distutils import dir_util
from datetime import datetime
from tabulate import tabulate
from json import JSONDecodeError


try:
    from hpogrid.components import validation
    from hpogrid.components.defaults import *
    from hpogrid.utils import stylus
except:
    raise ImportError('Cannot import hpogrid module. Try source setupenv first.')


kAction = ['create', 'recreate', 'update', 'list', 'show', 'remove']  
kConfigAction = ['create', 'recreate', 'update']
kConfigList = ['hpo_config', 'grid_config', 'search_space', 'model_config', 'project_config']

class ConfigurationBase():

    def __init__(self):
        self.initialize()
        parser = self.get_parser() 
        args = parser.parse_args(sys.argv[2:3])
        action = args.action
        parser = self.get_parser(action)
        args = parser.parse_args(sys.argv[3:])

        if action in kConfigAction:
            self.configure(args, action)
        elif hasattr(self, action):
            getattr(self, action)(**vars(args))
        else:
            print('Unrecognized action: {}'.format(action))
            parser.print_help()
            exit(1)            

    def initialize(self):
        self.description = 'Manage configuration'
        self.usage = 'hpogrid <config_type> <action> <config_name> [<options>]'
        self.config_type = 'SUPPRESS'
        self.list_columns = []
        self.show_columns = []

    def get_base_parser(self):
        parser = argparse.ArgumentParser(
            description=self.description,
            usage=self.usage) 
        return parser

    def get_parser(self, action=None):
        parser = self.get_base_parser()           
        if not action:
            parser.add_argument('action', help='Action to be performed', choices=kAction)    
        elif action == 'list':
            parser.add_argument('--expr', metavar='',
                help='Filter out config files that matches the expression')
        elif action == 'show':
            parser.add_argument('name', help='Name of config file to show')
        elif action == 'remove':
            parser.add_argument('name', help='Name of config file to remove')
        else:
            raise ValueError('Unknown method: {}'.format(action))
        return parser

    def get_base_dir(self, config_type=None, force_create=True):

        if config_type is None:
            config_type = self.config_type
            
        if config_type == 'project':
            base_dir = os.path.join(os.environ[kHPOGridEnvPath], 'project')
        else:
            base_dir = os.path.join(os.environ[kHPOGridEnvPath], 'config', config_type)

        if (not os.path.exists(base_dir)) and force_create:
            os.makedirs(base_dir, exist_ok=True)
        return base_dir

    def get_config_path(self, config_name=None, config_type=None, extension='.json'):
        if config_name is None:
            config_name = self.config_name
        if config_type is None:
            config_type = self.config_type
        base_dir = self.get_base_dir(config_type)
        config_base_name = '{}{}'.format(config_name, extension)
        config_path = os.path.join(base_dir, config_base_name)
        return config_path

    def remove(self, name):
        config_path = self.get_config_path(name)
        if os.path.exists(config_path):
            os.remove(config_path)
            print('INFO: Removed file {}'.format(config_path))
        else:
            print('ERROR: Cannot remove file {}. File does not exist.'.format(config_path))

    def get_updated_config(self, config):
        non_updated_keys = []
        for key in config:
            if config[key] is None:
                non_updated_keys.append(key)
        for key in non_updated_keys:
            config.pop(key, None)
        config_path = self.get_config_path(config['name'])
        if not os.path.exists(config_path):
            raise FileNotFoundError('Configuration file {} not found. Update aborted.'.format(config_path))
        old_config = json.load(open(config_path))
        config = {**old_config, **config}
        return config

    def _retain_only_updated_options(self):
        parser = self.get_parser('update')
        for action in parser._actions:
            if (len(action.option_strings) > 0) and (action.default != '==SUPPRESS=='):
                action.default=None
        args = parser.parse_args(sys.argv[3:])
        return args

    def configure(self, args, action='create'):
        if action == 'update':
            args = self._retain_only_updated_options()

        config = vars(args)
        
        if action == 'update':
            config = self.get_updated_config(config)

        self.process_config(config)
        if config is not None:
            self.save(config, args.name, action)
        return config

    def process_config(self, config):
        for key in config:
            if isinstance(config[key], bool):
                config[key] = int(config[key])
        return config

    def save(self, config, name, action='create'):

        config_path = self.get_config_path(name)
        if (os.path.exists(config_path)) and (action=='create'):
            print('ERROR: {} configuration with name {} already exists.'
                'If you want to overwrite, use "recreate" or "update" action instead of "create".'.format(
                self.config_type, name))
        else:
            with open(config_path, 'w') as config_file:
                json.dump(config, config_file, indent=2)
            action_map = { 'create': 'Created', 'recreate': 'Recreated', 'update': 'Updated'}
            print('INFO: {} {} configuration {}'.format(action_map[action], self.config_type, config_path))
            self.show(name)

    def get_config_list(self, expr=None):
        if not expr:
            expr = '*'
        base_dir = self.get_base_dir()
        config_list = [os.path.splitext(d)[0] for d in os.listdir(base_dir) if not d.startswith('.')]
        if expr is not None:
            config_list = fnmatch.filter(config_list, expr)
        return config_list

    def load_config(self, config_name):
        config_path = self.get_config_path(config_name)
        if not (os.path.exists(config_path)):
            raise FileNotFoundError("The configuration file {} does not exist.".format(config_path))
        config = json.load(open(config_path))
        return config

    def list(self, expr=None):
        config_list = self.get_config_list(expr)
        table = stylus.create_table(config_list, self.list_columns)
        print(table)

    def show(self, name):
        config = self.load_config(name)
        table = stylus.create_table(config.items(), self.show_columns)
        print(table)

class HPOConfiguration(ConfigurationBase):

    def initialize(self):
        self.description = 'Manage configuration for hyperparameter optimization'
        self.usage = 'hpogrid hpo_config <action> <config_name> [<options>]'
        self.config_type = 'hpo'
        self.list_columns = ['HPO Configuration']
        self.show_columns = ['Attribute', 'Value']    

    def get_parser(self, action=None):
        parser = self.get_base_parser()        
        if action in kConfigAction:
            parser.add_argument('name', help= "Name given to the configuration file")
            parser.add_argument('-a','--algorithm', 
                                help='Algorithm for hyperparameter optimization', 
                                default=kDefaultSearchAlgorithm, choices=kSearchAlgorithms)
            parser.add_argument('-m', '--metric', metavar='',
                                help='Evaluation metric to be optimized', 
                                default=kDefaultMetric)
            parser.add_argument('-o', '--mode', 
                                help='Mode of optimization (either "min" or "max")', 
                                default=kDefaultMode, choices=kMetricMode)
            parser.add_argument('-s','--scheduler', 
                                help='Trial scheduling method for hyperparameter optimization',
                                default=kDefaultScheduler, choices=kSchedulers)
            parser.add_argument('-t','--trials', metavar='',
                                help='Number of trials (search points)', 
                                type=int, default=kDefaultTrials)
            parser.add_argument('-l', '--log_dir', metavar='',
                                help='Logging directory',
                                default=kDefaultLogDir)
            parser.add_argument('-v','--verbose', action='store_true', 
                                help='Check to enable verbosity')
            parser.add_argument('--stop', metavar='',
                                help='Stopping criteria for the training',
                                default=kDefaultStopping)
            parser.add_argument('--scheduler_param', metavar='',
                                help='Extra parameters given to the trial scheduler', 
                                default=kDefaultSchedulerParam)
            parser.add_argument('--algorithm_param', metavar='',
                                help='Extra parameters given to the hyperparameter optimization algorithm',
                                default=kDefaultAlgorithmParam)
        else:
            parser = super().get_parser(action)
        return parser

    def process_config(self, config):
        super().process_config(config)

        json_interp = ['stop', 'scheduler_param', 'algorithm_param']
        for key in json_interp:
            if (key in config) and isinstance(config[key],str):
                try:
                    config[key] = json.loads(config[key])
                except JSONDecodeError:
                    print('ERROR: Cannot decode the value of {} into json format.'
                        'Please check your input.'.format(key))
                    return None                

        return config

    
class ValidateSites(argparse.Action):
    def __call__(self, parser, args, values, option_string=None):
        for site in values:
            if site not in kGPUGridSiteList:
                raise ValueError('Invalid site {}. '
                    'Please choose one of {}'.format(site, kGPUGridSiteList))
        setattr(args, self.dest, ','.join(values))
                
                
class GridConfiguration(ConfigurationBase):

    def initialize(self):
        self.description = 'Manage configuration for grid job submission'
        self.usage = 'hpogrid grid_config <action> <config_name> [<options>]'
        self.config_type = 'grid'
        self.list_columns = ['Grid Configuration']
        self.show_columns = ['Attribute', 'Value']  

    def get_parser(self, action=None):
        parser = self.get_base_parser()
        if action in kConfigAction:         
            parser.add_argument('name', help = "Name given to the configuration file")
            parser.add_argument('-s', '--site', 
                                help='Name of the grid site to where the jobs are submitted',
                                required=False, default=kDefaultGridSite,
                                action=ValidateSites)
            parser.add_argument('-c', '--container', metavar='',
                                help='Name of the docker or singularity container in which the jobs are run', 
                                required=False, default=kDefaultContainer)
            parser.add_argument('-r', '--retry',
                                help='Check to enable retrying faild jobs',
                                action='store_true')
            parser.add_argument('-i', '--inDS', metavar='',
                                help='Name of input dataset')
            parser.add_argument('-o', '--outDS', metavar='',
                                help='Name of output dataset', 
                                default=kDefaultOutDS)                                                                        
        else:
            parser = super().get_parser(action)
        return parser


class SearchSpaceConfiguration(ConfigurationBase):

    def initialize(self):
        self.description = 'Manage configuration for hyperparameter search space'
        self.usage = 'hpogrid search_space <action> <config_name> <search_space_definition>'
        self.config_type = 'search_space'
        self.list_columns = ['Search Space Configuration']
        self.show_columns = ['Attribute', 'Value']  


    def get_parser(self, action=None):
        parser = self.get_base_parser()               
        if action in kConfigAction:     
            parser.add_argument('name', 
                help='Name given to the configuration file')  
            parser.add_argument('search_space', 
                help='A json decodable string defining the search space')  
        else:
            parser = super().get_parser(action)
        return parser

    def configure(self, args, action='create'):

        config = vars(args)
        config_name = config.pop('name', None)
        search_space = config.pop('search_space', None)

        try:
            search_space = json.loads(search_space)
        except JSONDecodeError:
            print('ERROR: Cannot to decode input string into json format. Please check your input.')
            return None

        if action == 'update':
            config_path = self.get_config_path(config_name)
            if not os.path.exists(config_path):
                raise FileNotFoundError('Search space file {} not found. Update aborted.'.format(config_path))
            old_serach_space = json.load(open(config_path))
            search_space = {**old_serach_space, **search_space}
        
        if validation.validate_search_space(search_space):
            self.save(search_space, config_name, action)
        return search_space

class ModelConfiguration(ConfigurationBase):

    def initialize(self):
        self.description = 'Manage configuration for machine learning model'
        self.usage = 'hpogrid model_config <action> <config_name> [<options>]'
        self.config_type = 'model'
        self.list_columns = ['Model Configuration']
        self.show_columns = ['Attribute', 'Value']  

    def get_parser(self, action=None):
        parser = self.get_base_parser()              
        if action in kConfigAction:  
            parser.add_argument('name', help= "Name given to the configuration file")            
            parser.add_argument('-s','--script', metavar='',
                help='Name of the training script where the function or class that defines'
                     ' the training model will be called to perform the training')
            parser.add_argument('-m','--model', metavar='',
                help='Name of the function or class that defines the training model')        
            parser.add_argument('-p','--param', metavar='',
                help='Extra parameters to be passed to the training model',
                default=kDefaultModelParam)
        else:
            parser = super().get_parser(action)
        return parser

    def process_config(self, config):
        super().process_config(config)

        if ('param' in config) and isinstance(config['param'],str):
            try:
                config['param'] = json.loads(config['param'])
            except JSONDecodeError:
                print('ERROR: Cannot decode input param into json format. Please check your input.')
                return None

        return config

class ProjectConfiguration(ConfigurationBase):

    def initialize(self):
        self.description = 'Manage a project for hyperparamter optimization'
        self.usage = 'hpogrid project <action> <project_name> [<options>]'
        self.config_type = 'project'
        self.list_columns = ['Project Title']
        self.show_columns = ['Attribute', 'Value']  
        self.project_config = {}


    def get_parser(self, action=None):
        parser = self.get_base_parser()           
        if action in kConfigAction:          
            parser.add_argument('name', help= "Name given the project")
            parser.add_argument('-p','--scripts_path', metavar='',
                help='Path to where the training scripts'
                ' (or the directory containing the training scripts) are located')
            parser.add_argument('-o','--hpo_config', metavar='',
                help='Name of the hpo configuration to use')
            parser.add_argument('-g','--grid_config', metavar='',
                help='Name of the grid configuration to use')
            parser.add_argument('-m','--model_config', metavar='',
                help='Name of the model configuration to use')
            parser.add_argument('-s','--search_space', metavar='',
                help='Name of the search space configuration to use')
        else:
            parser = super().get_parser(action)
        return parser

    def get_updated_config(self, config):
        return config

    def process_config(self, config):
        self.project_config = config

        print('INFO: Checking validity of input paths...')
        # check if path to training scripts exists
        if (config['scripts_path'] is not None):
            scripts_path = config['scripts_path']
            if not os.path.exists(scripts_path):
                print('ERROR: Path to training scripts {} does not exist.'
                       'Copy to project will be skipped.'.format(scripts_path))
                config['scripts_path'] = None
        else:
            print('INFO: Path to training scripts is not specified. Checking will be skipped.')

        config_type_map = {
            'hpo_config': 'hpo',
            'grid_config': 'grid',
            'model_config': 'model',
            'search_space': 'search_space'
        }

        # check if input configuration files exist
        for key in config_type_map:
            if (key in config) and (config[key] is not None):
                config_type = config_type_map[key]
                config_path = self.get_config_path(config[key], config_type)
                if not os.path.exists(config_path):
                    print('WARNING: Path to {} config {} does not exist.'.format(config_type, config_path))
                    config[key] = None
                else:
                    config[key] = config_path
            else:
                print('INFO: Path to {} config is not specified. Checking will be skipped.'.format(config_type_map[key]))
        print('INFO: Successfully validated input paths.')

        return config

    def get_project_path(self, proj_name):
        return self.get_config_path(proj_name, self.config_type, extension='')

    def save(self, config, name, action='create'):
        proj_name = name
        proj_path = self.get_project_path(proj_name)
        if (os.path.exists(proj_path)):
            if  action == 'create':
                print('ERROR: Project titled {} already exists. If you want to overwrite,'
                    ' use "recreate" or "update" action instead of "create".'.format(proj_name))
                return None
            elif action == 'recreate':
                backup_dir = self.get_config_path('backup', self.config_type, extension='')
                os.makedirs(backup_dir, exist_ok=True)
                timestamp = datetime.now().strftime("%Y%m%d%H%M%S")
                backup_proj_name = os.path.join(backup_dir, '{}_{}'.format(proj_name, timestamp))
                shutil.move(proj_path, backup_proj_name)
                print('INFO: Recreating project. Original project moved to backup directory {}.'.format(
                    backup_proj_name))
        # create project directories
        scripts_dir = os.path.join(proj_path, 'scripts')
        config_dir = os.path.join(proj_path, 'config')
        os.makedirs(proj_path, exist_ok=True)        
        os.makedirs(scripts_dir, exist_ok=True)
        os.makedirs(config_dir, exist_ok=True)

        # copy input conifigurations
        print('INFO: Copying input configurations to the project directory')       
        if ('scripts_path' in config) and (config['scripts_path'] is not None):
            # copy contents of directory to project/scrsipts/
            if os.path.isdir(config['scripts_path']):
                dir_util.copy_tree(config['scripts_path'], scripts_dir)
            else:
                shutil.copy2(config['scripts_path'], scripts_dir)
            print('INFO: From {} copied training scripts to {}'.format(config['scripts_path'], scripts_dir))

        for key in kConfigList:
            if (key in config) and (config[key] is not None):
                dest = os.path.join(config_dir, '{}.json'.format(key))
                shutil.copy2(config[key], dest)
                print('INFO: Copied {} to {}'.format(config[key], dest))

        project_config = {}
        project_config_path = os.path.join(config_dir, kProjectConfigName)
        if os.path.exists(project_config_path):
            project_config = json.load(open(project_config_path))
        
        project_config = {**self.project_config, **project_config}
        with open(project_config_path,'w') as proj_config_file:
            json.dump(project_config, proj_config_file, indent=2)

    def get_config_list(self, expr=None):
        project_list = [ s for s in super().get_config_list() if s is not 'backup']
        return project_list

    def remove(self, name):
        proj_path = self.get_project_path(name)
        if os.path.exists(proj_path):
            print('WARNING: To avoid accidental deletion of important files. '
                'Please delete your project manually at:\n{}'.format(proj_path))
        else:
            print('ERROR: Cannot remove project in {}. Path does not exist.'.format(proj_path))

    def show(self, name):
        proj_path = self.get_project_path(name)
        config_path = os.path.join(proj_path, 'config', kProjectConfigName)
        if os.path.exists(config_path):
            with open(config_path,'r') as config_file:
                config = json.load(config_file)
            table = stylus.create_table(config.items(), self.show_columns)
            print(table)
        else:
            print('ERROR: Project {} does not exist.'.format(name))


from abc import ABC
from typing import List, Union
import argparse
import copy
from functools import partial

from hpogrid import algorithm as algo
from hpogrid import search_space as space

class Generator(ABC):
    def __init__(self, search_space, metric, mode, **args):
        self.evaluated_points = []
        self.metric = metric
        self.mode = mode
        self.search_space = search_space
        if mode == "max":
            self.signature = -1.
        elif mode == "min":
            self.signature = 1.
        self.searcher = self.get_searcher(search_space, metric, mode, **args)

    @abstractmethod    
    def get_searcher(self, search_space, metric, mode, **args):
        pass

    @abstractmethod    
    def tell(self, point, result):
        pass

    def feed(self, points:List, results:List, redo=False):
        if redo:
            self.clear_history()
        for point, result in zip(points, results):
            self.tell(point, result)

    @abstractmethod    
    def ask(self, n_points:int =None):
        pass

    @abstractmethod  
    def clear_history(self):
        pass


class NeverGradGenerator(Generator):
    def __init__(self, search_space, metric, mode):
        self.reset()

    def get_searcher(self, search_space, metric, mode, **args):
        from algo.nevergrad_algorithm import NeverGradAlgoWrapper
        return NeverGradAlgoWrapper(metric, mode, search_space, **args)

    def tell(self, point, value):


class HyperOptGenerator(Generator):
    def get_searcher(self, search_space, metric, mode, **args):
        import hyperopt as hpo
        from hyperopt.fmin import generate_trials_to_calculate
        searcher = hpo.tpe.suggest
        if 'gamma' in args:
            searcher = partial(searcher, gamma=gamma)
        from hpogrid.search_space.hyperopt_space import HyperOptSpace
        hyperopt_space = HyperOptSpace(search_space).get_search_space()
        self.domain = hpo.Domain(lambda spc: spc, hyperopt_space)
        self.trials = hpo.Trials()
        self.rstate = np.random.RandomState()
        return searcher


    def ask(self, n_points:int =None):
        points = []
        trial_ids = self.trials.new_trial_ids(n_points)
        self.trials.refresh()
        new_trials = self.searcher(trial_ids, self.domain, self.trials,
            self.rstate.randint(2**31 - 1))
        for trial in new_trials:
            config = hpo.base.spec_from_misc(trial["misc"])
            memo = self.domain.memo_from_config(config)
            point = hpo.pyll.rec_eval(self.domain.expr, memo=memo)
            points.append(copy.deepcopy(point))
        return points

    def clear_history(self):
        self.trials.delete_all()

    def tell(self, point, value):
        fake_trial = generate_trials_to_calculate([point])
        fake_trial.refresh()
        trial = fake_trial._trials[0]
        trial['state'] = hpo.base.JOB_STATE_DONE
        trial['result'] = {"loss": self.signature * value, "status": "ok"}
        self.trials.insert_trial_doc(trial)
        self.trials.refresh()

        
class SkOptGenerator(Generator):
    def get_searcher(self, search_space, metric, mode, **args):
        from ax.service.ax_client import AxClient
        ax_client = AxClient(enforce_sequential_optimization=False, verbose_logging=False)
        from hpogrid.search_space.ax_space import AxSpace
        search_space = AxSpace(search_space).get_search_space()
        if mode == 'max':
            minimize = False
        elif mode == 'min':
            minimize = True
        else:
            raise ValueError('mode of evaluation metric can only be "min" or "max"')

        ax_client.create_experiment(
            parameters=search_space,
            objective_name=metric,
            minimize=minimize)
        return ax_client    

class AxGenerator(Generator):
    def get_searcher(self, search_space, metric, mode):
        from ax.service.ax_client import AxClient
        ax_client = AxClient(enforce_sequential_optimization=False, verbose_logging=False)
        from hpogrid.search_space.ax_space import AxSpace
        search_space = AxSpace(search_space).get_search_space()
        if mode == 'max':
            minimize = False
        elif mode == 'min':
            minimize = True
        else:
            raise ValueError('mode of evaluation metric can only be "min" or "max"')

        ax_client.create_experiment(
            parameters=search_space,
            objective_name=metric,
            minimize=minimize)
        return ax_client

    def tell(self, point, value):
        _, trial_index = self.searcher.attach_trial(point)
        metric_dict = { self.metric: (value, 0.0)}
        self.searcher.complete_trial(
            trial_index=trial_index, raw_data=metric_dict)

    def ask(self, n_points:int =None):
        points = []
        for _ in range(n_points):
            point, trial_index = self.searcher.get_next_trial()
            trial = self.searcher._get_trial(trial_index=trial_index)
            trial.mark_abandoned()
            points.append(point)
        return points

    def clear_history(self):
        self.searcher = self.get_searcher(self.search_space, self.metric, self.mode)


"""
    Module for formatting tables, texts and figures used in hpogrid
"""
from tabulate import tabulate
import pandas as pd
import yaml
import copy
from pdb import set_trace

kDefaultTableStyle = 'psql'
kDefaultStrAlign = 'left'

class ColorCode():
    RED = '\033[0;91m'
    GREEN = '\033[92m'
    YELLOW = '\033[93m'
    BLUE = '\033[94m'
    MAGENTA = '\033[95m'
    CYAN = '\033[96m'
    ENDC = '\033[0m'
    BOLD = '\033[1m'
    UNDERLINE = '\033[4m'
    BLINK = '\033[5m'

def join_options(options):
    return ' '.join(["--{} {}".format(key,value) for (key,value) in options.items()])

def create_table(data, columns=None, indexed=True, transpose=False,
    tableformat=kDefaultTableStyle, stralign=kDefaultStrAlign,
    wrap_nested_dict=False):
    _data = copy.deepcopy(data)
    if isinstance(_data, dict):
        for key in _data:
            if isinstance(_data[key], dict):
                _data[key] = yaml.dump(_data[key], allow_unicode=True,
                                       default_flow_style=False, sort_keys=False)
    _data = _data.items() if isinstance(data, dict) else _data
    df = pd.DataFrame(_data, columns=columns)
    if transpose:
        df = df.transpose()
    table = tabulate(df, showindex=indexed, headers=df.columns, 
        tablefmt=tableformat,
        stralign=stralign)
    return table


import re
import os
import sys
import shutil
import atexit
import argparse
from pandatools.Group_argparse import GroupArgParser
try:
    from urllib import quote
except ImportError:
    from urllib.parse import quote
import types
import json
import copy
from pandatools.MiscUtils import commands_get_output, commands_get_status_output
from pandatools import MiscUtils
    
    
from pandatools import Client
from pandatools import PsubUtils
from pandatools import AthenaUtils
from pandatools import PLogger



class PrunWrapper():
    def __init__(self):
        taskParamMap = {}
        taskParamMap['osInfo'] = PsubUtils.get_os_information()
    
    def to_json(self):
        pass
    
    def to_xml(self):
        pass






import sys, os
import argparse
import json

from hpogrid.components.defaults import *
from hpogrid.components import validation
from hpogrid.configuration.configuration_base import ConfigurationBase, kConfigAction

class SearchSpaceConfiguration(ConfigurationBase):

    def initialize(self):
        self.description = 'Manage configuration for hyperparameter search space'
        self.usage = 'hpogrid search_space <action> <config_name> <search_space_definition>'
        self.config_type = 'search_space'
        self.list_columns = ['Search Space Configuration']
        self.show_columns = ['Hyperparameters', 'Search Space']  


    def get_parser(self, action=None):
        parser = self.get_base_parser()               
        if action in kConfigAction:     
            parser.add_argument('name', 
                help='Name given to the configuration file')  
            parser.add_argument('search_space', 
                help='A json decodable string defining the search space')  
        else:
            parser = super().get_parser(action)
        return parser

    def configure(self, args, action='create'):

        config = vars(args)
        config_name = config.pop('name', None)
        search_space = config.pop('search_space', None)

        try:
            search_space = json.loads(search_space)
        except JSONDecodeError:
            print('ERROR: Cannot to decode input string into json format. Please check your input.')
            return None

        if action == 'update':
            config_path = self.get_config_path(config_name)
            if not os.path.exists(config_path):
                raise FileNotFoundError('Search space file {} not found. Update aborted.'.format(config_path))
            old_serach_space = json.load(open(config_path))
            search_space = {**old_serach_space, **search_space}
        
        if validation.validate_search_space(search_space):
            self.save(search_space, config_name, action)
        return search_space


import skopt
from skopt.space import Real, Categorical, Integer

from hpogrid.search_space.base_space import BaseSpace

class SkOptSpace(BaseSpace):

	def __init__(self, search_space = None):
		self.library = 'skopt'
		super().__init__(search_space)	
		

	def reset_space(self):
		self.search_space = []

	def append(self, space_value):
		self.search_space.append(space_value)

	def categorical(self, label, categories, grid_search = False):
		if grid_search != False:
			raise ValueError('{} does not allow grid search'.format(self.library))
		return Categorical(name=label, categories=categories)

	def uniformint(self, label, low, high):
		return Integer(name=label, low=low, high=high, prior="uniform")

	def uniform(self, label, low, high):
		return Real(name=label, low=low, high=high, prior='uniform')

	def loguniform(self, label, low, high, base=10):
		return Real(name=label, low=low, high=high, prior='log-uniform', base=base)

	def loguniformint(self, label, low, high, base=10):
		return Integer(name=label, low=low, high=high, prior="log-uniform", base=base)

	def fixed(self, label, value):
		return self.categorical(label=label, categories=[value])


import ax

from hpogrid.search_space.base_space import BaseSpace

class AxSpace(BaseSpace):

	def __init__(self, search_space = None):
		self.library = 'ax'
		super().__init__(search_space)	
		

	def reset_space(self):
		self.search_space = []

	def append(self, space_value):
		self.search_space.append(space_value)
	
	def _range(self, name, bounds, value_type = "bounds", log_scale = False):
		space_value = {
			"name": name,
			"type": "range",
			"bounds": bounds,
			"value_type": value_type,
			"log_scale": log_scale
		}
		return space_value


	def categorical(self, label, categories, grid_search = False):
		space_value = {
			"name": label,
			"type": "choice",
			"values": categories
		}
		return space_value

	def uniform(self, label, low, high):
		return self._range(label, [low, high], "float")

	def uniformint(self, label, low, high):
		return self._range(label, [low, high], "int")

	def loguniform(self, label, low, high, base = 10):
		if base != 10:
			raise ValueError('{} search space only allows base 10 for loguniform sampling'.format(self.library))
		return self._range(label, [low, high], "float", log_scale=True)

	def qloguniform(self, label, low, high, q = 1):
		if q != 1:
			raise ValueError('{} search space only allows q = 1 for qloguniform sampling'.format(self.library))
		return self._range(label, [low, high], "int", log_scale=True)


	def fixed(name, label, value):
		space_value = {
			"name": label,
			"value": value
		}
		return space_value


import ray

from ray.tune.schedulers import AsyncHyperBandScheduler

# use with
# ax
# tune
# hyperopt
class AsyncHyperBandSchedulerWrapper():
    def __init__(self):
        self.scheduler = None
    def create(self, metric, mode, **args):
        self.scheduler = AsyncHyperBandScheduler(metric=metric, mode=mode, **args)
        return self.scheduler



import numpy as np 
import json
from pdb import set_trace

class BaseSpace():

	test_space = {
		'activation' : {
			'method': 'categorical',
			'dimension': {
				'categories': ['relu', 'sigmoid'],
				'grid_search' : 0
			}
		},
		'learning_rate' : {
			'method': 'loguniform',
			'dimension': {
				'low': 1e-6,
				'high': 1e-1,
				'base': 10
			}
		},
		'batchsize': {
			'method': 'categorical',
			'dimension': {
				'categories': [64,128,256,512,1024,2048],
				'grid_search' : 0
			}
		},
		'max_depth': {
			'method': 'uniformint',
			'dimension': {
				'low': 1,
				'high': 30
			}
		},
		'subsample': {
			'method': 'uniform',
			'dimension': {
				'low': 0.1,
				'high': 1.0
			}
		}
	}

	test_space_base_e = {
		'activation' : {
			'method': 'categorical',
			'dimension': {
				'categories': ['relu', 'sigmoid'],
				'grid_search' : 0
			}
		},
		'learning_rate' : {
			'method': 'loguniform',
			'dimension': {
				'low': np.e**-11,
				'high': np.e**-6,
				'base': np.e
			}
		},
		'batchsize': {
			'method': 'categorical',
			'dimension': {
				'categories': [64,128,256,512,1024,2048],
				'grid_search' : 0
			}
		},
		'max_depth': {
			'method': 'uniformint',
			'dimension': {
				'low': 1,
				'high': 30
			}
		},
		'subsample': {
			'method': 'uniform',
			'dimension': {
				'low': 0.1,
				'high': 1.0
			}
		}
	}	

	def __init__(self, search_space = None):
		self.reset_space()
		if search_space is not None:
			self.create(search_space)


	def get_search_space(self):
		return self.search_space
	
	def reset_space(self):
		self.library = 'base'
		self.search_space = {}

	def _method_error(self, method):
		raise ValueError('the method {} is not supported in {}'.format(method, self.library))

	def import_json(file):
		self.search_space = json.load(open(file))		
		
	def create(self,search_space = None):
		if search_space is None:
			search_space = self.search_space
		elif isinstance(search_space, str):
			search_space = json.loads(search_space)

		methods_map = {
			'categorical': self.categorical,
			'uniform': self.uniform,
			'uniformint': self.uniformint,
			'quniform': self.quniform,
			'loguniform': self.loguniform,
			'qloguniform': self.qloguniform,
			'normal': self.normal,
			'qnormal': self.qnormal,
			'lognormal': self.lognormal,
			'qlognormal': self.qlognormal,
			'fixed': self.fixed,
		}

		for hp in search_space:
			method = search_space[hp]['method']
			if method not in methods_map:
				self._method_error(method)
			space_value = methods_map[method](label = hp, **search_space[hp]['dimension'])
			if space_value is None:
				self._method_error(method)
			self.append(space_value)
		return self.search_space


	def append(self, space_value):
		self.search_space[space_value[0]] = space_value[1]

	def categorical(self, label, categories, grid_search = False):
		'''
		returns one of the categories, which should be a list
		'''
		space_value = (label, {'method': 'categorical', 'dimension': categories})
		return space_value

	def uniformint(self, label, low, high):
		'''
		returns a random integer in the range [low, high).
		'''
		space_value = (label, {'method': 'uniformint', 
			'dimension': { 'low': low, 'high': high}})
		return space_value

	def uniform(self, label, low, high):
		'''
		returns a value uniformly distributed between low and high.
		'''
		space_value = (label, {'method': 'uniform', 
			'dimension': { 'low': low, 'high': high}})
		return space_value

	def quniform(self, label, low, high, q):
		'''
		returns a value like round(uniform(low, high) / q) * q
		'''
		space_value = (label, {'method': 'quniform', 
			'dimension': { 'low': low, 'high': high, 'q': q}})
		return space_value

	def loguniform(self, label, low, high, base):
		'''
		loguniform returns a value drawn according to power(uniform(low, high), base = base) 
		so that the logarithm of the return value is uniformly distributed
        '''
		space_value = (label, {'method': 'loguniform', 
			'dimension': { 'low': low, 'high': high, 'base': base}})
		return space_value

	def qloguniform(self, label, low, high, q, base):
		'''
		Returns a value like round(exp(uniform(low, high)) / q) * q
		'''
		space_value = (label, {'method': 'qloguniform', 
			'dimension': { 'low': low, 'high': high, 'q': q, 'base': base}})
		return space_value

	def normal(self, label, mu, sigma):
		'''
		returns a real value that's normally-distributed with mean mu and standard deviation sigma
		'''
		space_value = (label, {'method': 'normal', 
			'dimension': { 'mu': mu, 'sigma': sigma}})
		return space_value


	def lognormal(self, label, mu, sigma, base):
		'''
		returns a value drawn according to power(normal(mu, sigma), base) so that the logarithm of the 
		return value is normally distributed
		'''
		space_value = (label, {'method': 'lognormal', 
			'dimension': { 'mu': mu, 'sigma': sigma, 'base': base}})
		return space_value

	def qnormal(self, label, mu, sigma, q):
		'''
		returns a value like round(normal(mu, sigma) / q) * q
		'''
		space_value = (label, {'method': 'qnormal', 
			'dimension': { 'mu': mu, 'sigma': sigma, 'q': q}})
		return space_value

	def qlognormal(self, label, mu, sigma, q, base):
		'''
		returns a value like round(power(normal(mu, sigma), base) / q) * q
		'''
		space_value = (label, {'method': 'qlognormal', 
			'dimension': { 'mu': mu, 'sigma': sigma, 'q': q, 'base': base}})
		return space_value

	def fixed(self, value):
		'''
		returns a fixed value
		'''
		space_value = (label, {'method': 'fixed', 'dimension': value})
		return space_value


import sys, os
import argparse
import json
import fnmatch

from hpogrid.utils import stylus
from hpogrid.components.defaults import *

kActionList = ['create', 'recreate', 'update', 'list', 'show', 'remove']  
kConfigAction = ['create', 'recreate', 'update']

class ConfigurationBase():

    def __init__(self):
        self.initialize()
        parser = self.get_parser() 
        args = parser.parse_args(sys.argv[2:3])
        action = args.action
        parser = self.get_parser(action)
        args = parser.parse_args(sys.argv[3:])

        if action in kConfigAction:
            self.configure(args, action)
        elif hasattr(self, action):
            getattr(self, action)(**vars(args))
        else:
            print('Unrecognized action: {}'.format(action))
            parser.print_help()
            exit(1)            

    def initialize(self):
        self.description = 'Manage configuration'
        self.usage = 'hpogrid <config_type> <action> <config_name> [<options>]'
        self.config_type = 'SUPPRESS'
        self.list_columns = []
        self.show_columns = []

    def get_base_parser(self):
        parser = argparse.ArgumentParser(
            description=self.description,
            usage=self.usage) 
        return parser

    def get_parser(self, action=None):
        parser = self.get_base_parser()           
        if not action:
            parser.add_argument('action', help='Action to be performed', choices=kActionList)    
        elif action == 'list':
            parser.add_argument('--expr', metavar='',
                help='Filter out config files that matches the expression')
        elif action == 'show':
            parser.add_argument('name', help='Name of config file to show')
        elif action == 'remove':
            parser.add_argument('name', help='Name of config file to remove')
        else:
            raise ValueError('Unknown method: {}'.format(action))
        return parser

    def get_base_dir(self, config_type=None, force_create=True):

        if config_type is None:
            config_type = self.config_type
            
        if config_type == 'project':
            base_dir = os.path.join(os.environ[kHPOGridEnvPath], 'project')
        else:
            base_dir = os.path.join(os.environ[kHPOGridEnvPath], 'config', config_type)

        if (not os.path.exists(base_dir)) and force_create:
            os.makedirs(base_dir, exist_ok=True)
        return base_dir

    def get_config_path(self, config_name=None, config_type=None, extension='.json'):
        if config_name is None:
            config_name = self.config_name
        if config_type is None:
            config_type = self.config_type
        base_dir = self.get_base_dir(config_type)
        config_base_name = '{}{}'.format(config_name, extension)
        config_path = os.path.join(base_dir, config_base_name)
        return config_path

    def remove(self, name):
        config_path = self.get_config_path(name)
        if os.path.exists(config_path):
            os.remove(config_path)
            print('INFO: Removed file {}'.format(config_path))
        else:
            print('ERROR: Cannot remove file {}. File does not exist.'.format(config_path))

    def get_updated_config(self, config):
        non_updated_keys = []
        for key in config:
            if config[key] is None:
                non_updated_keys.append(key)
        for key in non_updated_keys:
            config.pop(key, None)
        config_path = self.get_config_path(config['name'])
        if not os.path.exists(config_path):
            raise FileNotFoundError('Configuration file {} not found. Update aborted.'.format(config_path))
        old_config = json.load(open(config_path))
        config = {**old_config, **config}
        return config

    def _retain_only_updated_options(self):
        parser = self.get_parser('update')
        for action in parser._actions:
            if (len(action.option_strings) > 0) and (action.default != '==SUPPRESS=='):
                action.default=None
        args = parser.parse_args(sys.argv[3:])
        return args

    def configure(self, args, action='create'):
        if action == 'update':
            args = self._retain_only_updated_options()

        config = vars(args)
        
        if action == 'update':
            config = self.get_updated_config(config)

        self.process_config(config)
        config_name = config.pop('name', None)
        if config is not None:
            self.save(config, config_name, action)
        return config

    def process_config(self, config):
        for key in config:
            if isinstance(config[key], bool):
                config[key] = int(config[key])
        return config

    def save(self, config, name, action='create'):

        config_path = self.get_config_path(name)
        if (os.path.exists(config_path)) and (action=='create'):
            print('ERROR: {} configuration with name {} already exists.'
                'If you want to overwrite, use "recreate" or "update" action instead of "create".'.format(
                self.config_type, name))
        else:
            with open(config_path, 'w') as config_file:
                json.dump(config, config_file, indent=2)
            action_map = { 'create': 'Created', 'recreate': 'Recreated', 'update': 'Updated'}
            print('INFO: {} {} configuration {}'.format(action_map[action], self.config_type, config_path))
            self.show(name)

    def get_config_list(self, expr=None):
        if not expr:
            expr = '*'
        base_dir = self.get_base_dir()
        config_list = [os.path.splitext(d)[0] for d in os.listdir(base_dir) if not d.startswith('.')]
        if expr is not None:
            config_list = fnmatch.filter(config_list, expr)
        return config_list

    def load_config(self, config_name):
        config_path = self.get_config_path(config_name)
        if not (os.path.exists(config_path)):
            raise FileNotFoundError("The configuration file {} does not exist.".format(config_path))
        config = json.load(open(config_path))
        return config

    def list(self, expr=None):
        config_list = self.get_config_list(expr)
        table = stylus.create_table(config_list, self.list_columns)
        print(table)

    def show(self, name):
        config = self.load_config(name)
        formatted_str = stylus.create_table(config, self.show_columns, indexed=False,
                                           wrap_nested_dict=True)
        print(formatted_str)



from abc import ABC, abstractmethod
import copy
import json
from typing import List, Dict, Union
from functools import partial

from hpogrid.utils import stylus

import argparse

class Generator(ABC):
    def __init__(self, search_space:Dict, metric:str, mode:str, **args):
        self.evaluated_points = []
        self.metric = metric
        self.mode = mode
        self.search_space = search_space
        if mode == "max":
            self.signature = -1.
        elif mode == "min":
            self.signature = 1.
        self.searcher = self.get_searcher(search_space, metric, mode, **args)

    @abstractmethod    
    def get_searcher(self, search_space:Dict, metric:str, mode:str, **args):
        pass

    @abstractmethod    
    def ask(self, n_points:int =1):
        pass

    @abstractmethod    
    def tell(self, point:Dict, result):
        pass

    def feed(self, points:List, results:List, redo=False):
        assert len(points)==len(results), 'points and results must have the same length'
        if redo:
            self.clear_history()
        for point, result in zip(points, results):
            self.tell(point, result)

    def clear_history(self):
        self.searcher = self.get_searcher(self.search_space, self.metric, self.mode)


    def _to_metric_values(self, results:Union[List, Dict, float, int]):

        def extract_from_dict(data):
            if self.metric in data:
                return data[self.metric]
            else:
                raise KeyError('result should contain value for'
                ' the metric: {}'.format(self.metric))
        if isinstance(results, List):
            if all(isinstance(r, Dict) for r in results):
                return [extract_from_dict(r) for r in results]
            elif all(isinstance(r, (int, float)) for r in results):
                return results
            else:
                raise ValueError('invalid result format.')
        elif isinstance(results, Dict):
            return extract_from_dict(results)
        elif isinstance(results, (int, float)):
            return results
            
    def show(self, points:List, results:Union[List, None]=None):
        if results:
            assert len(points)==len(results), 'points and results must have the same length'
            metric_values = self._to_metric_values(results)
            data = [ {**p, self.metric:r } for p,r in zip(points, results)]
        else:
            data = points
        table = stylus.create_table(data)
        print(table)        


import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="hpogrid", # Replace with your own username
    version="0.2.9",
    author="Alkaid Cheng",
    author_email="chi.lung.cheng@cern.ch",
    description="A workflow for hyperparamter optimization using the ATLAS grid resources",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://gitlab.cern.ch/aml/hyperparameter-optimization/alkaid-qt/hpogrid",
    packages=setuptools.find_packages(),
    classifiers=[
        "Programming Language :: Python :: 3.5",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    install_requires=[   
          'ray[tune]',
          'pandas',
          'tqdm',
          'tabulate',
          'numpy',
          'psutil'
      ],
    scripts=['bin/hpogrid'],
    python_requires='>=3.6',
)



import numpy as np

import ConfigSpace as CS
import ConfigSpace.hyperparameters as CSH


from hpogrid.search_space.base_space import BaseSpace

class BOHBSpace(BaseSpace):

	def __init__(self, search_space = None):
		self.library = 'bohb'
		super().__init__(search_space)	
		

	def reset_space(self):
		self.search_space = CS.ConfigurationSpace(seed=1)	

	def append(self, space_value):
		self.search_space.add_hyperparameter(space_value)

	def categorical(self, label, categories, grid_search = False):
		if grid_search != False:
			raise ValueError('{} does not allow grid search'.format(self.library))
		return CSH.CategoricalHyperparameter(name=label, choices=categories)

	def uniformint(self, label, low, high):
		return CSH.UniformIntegerHyperparameter(name=label, lower=low, upper=high)

	def uniform(self, label, low, high):
		return CSH.UniformFloatHyperparameter(name=label, lower=low, upper=high)

	def quniform(self, label, low, high, q):
		return CSH.UniformFloatHyperparameter(name=label, lower=low, upper=high, q=q)

	def qloguniform(self, label, low, high, q, base=np.e):
		if base != np.e:
			raise ValueError('{} only allows base e for qloguniform'.format(self.library))
		return CSH.UniformFloatHyperparameter(name=label, lower=low, upper=high, q=q, log=True)

	def loguniform(self, label, low, high, base=np.e):
		if base != np.e:
			raise ValueError('{} only allows base e for loguniform'.format(self.library))
		return CSH.UniformFloatHyperparameter(name=label, lower=low, upper=high, log=True)


	def normal(self, label, mu, sigma):
		return CSH.NormalFloatHyperparameter(name=label, mu=mu, sigma=sigma)

	def qnormal(self, label, mu, sigma, q):
		return CSH.NormalFloatHyperparameter(name=label, mu=mu, sigma=sigma, q=q)

	def qlognormal(self, label, mu, sigma, q, base=np.e):
		if base != np.e:
			raise ValueError('{} only allows base e for qlognormal'.format(self.library))
		return CSH.NormalFloatHyperparameter(name=label, mu=mu, sigma=sigma, q=q, log=True)

	def fixed(self, label, value):
		return self.CategoricalHyperparameter(name=label, choices=[value])



import os
import sys
import glob
import argparse
import json

from hpogrid.components.defaults import *
from hpogrid.utils import helper, stylus

class GridHandler():
    def __init__(self):

        # submit grid job via hpogrid executable
        if len(sys.argv) > 1:
            self.run_parser()

    def get_parser(self):
        parser = argparse.ArgumentParser(
                    formatter_class=argparse.RawDescriptionHelpFormatter)
        parser.add_argument('proj_name', help='the project to submit a grid job')               
        parser.add_argument('-n','--n_jobs', type=int, help='number of jobs to submit',
            default=1)
        parser.add_argument('-s','--site', help='site to submit the job to '
            '(this will override the grid config site setting)', choices=kGPUGridSiteList)
        return parser

    def run_parser(self):
        parser = self.get_parser()
        if os.path.basename(sys.argv[0]) == 'hpogrid':
            args = parser.parse_args(sys.argv[2:])
        else:
            args = parser.parse_args(sys.argv[1:])
        GridHandler.submit_job(args.proj_name, args.n_jobs, args.site)
                
    @staticmethod
    def submit_job(proj_name, n_jobs=1, site=None):

        grid_config = helper.get_project_config(proj_name)['grid_config']

        options = {}

        options['containerImage'] = grid_config['container']

        options['exec'] = '"hpogrid grid_run {}"'.format(proj_name)

        if not grid_config['retry']:
            options['disableAutoRetry'] = ''

        extra = {'forceStaged':'', 
                 'useSandbox': '',
                 'noBuild': ''}
                 
        options.update(extra)
        options['ctrWorkdir'] = kWorkDir
        options['ctrDatadir'] = kDataDir

        if grid_config['inDS']:
            options['inDS'] = grid_config['inDS']

        if '{HPO_PROJECT_NAME}' in grid_config['outDS']:
            grid_config['outDS'] = grid_config['outDS'].format(HPO_PROJECT_NAME=proj_name)
        options['outDS'] = grid_config['outDS']

        if not site:
            site = grid_config['site']
        if (site != 'ANY'):
            options['site'] = site
            if 'GPU' in site:
                options['cmtConfig'] = 'nvidia-gpu'
            else:
                options['nCore'] = '8'
        
        # options['workDir'] = project_path # does not really work
        project_path = helper.get_project_path(proj_name)

        # construct prun command
        command = stylus.join_options(options)

        # switch to working directory to send files to WNs
        with helper.cd(project_path):
            # submit grid jobs
            for _ in range(n_jobs):
                os.system("prun {}".format(command))



import os
import sys
import glob
import argparse
import json
from typing import Optional

from hpogrid.components.defaults import *
from hpogrid.utils import helper, stylus

class iDDSHandler():
    def __init__(self):

        # submit grid job via hpogrid executable
        if len(sys.argv) > 1:
            self.run_parser()

    def get_parser(self):
        parser = argparse.ArgumentParser(
                    formatter_class=argparse.RawDescriptionHelpFormatter)
        parser.add_argument('proj_name', help='the project to submit a grid job')               
        parser.add_argument('-s','--site', help='site to submit the job to '
            '(this will override the grid config site setting)', choices=kGPUGridSiteList)
        return parser

    def run_parser(self):
        parser = self.get_parser()
        if os.path.basename(sys.argv[0]) == 'hpogrid':
            args = parser.parse_args(sys.argv[2:])
        else:
            args = parser.parse_args(sys.argv[1:])
        iDDSHandler.submit_job(args.proj_name, args.site)
        
    @staticmethod
    def submit_job(project_name:str, site:Optional[str]=None):
        grid_config = helper.get_project_config(project_name)['grid_config']
        project_path = helper.get_project_path(project_name)
        idds_config_path = os.path.join(project_path, 'idds', kiDDSConfigName)
        idds_search_space_path = os.path.join(project_path, 'idds', kiDDSSearchSpaceName)
        options = {}
        options['loadJson'] = idds_config_path 
        options['searchSpaceFile'] = idds_search_space_path
        if not site:
            site = grid_config['site']
        if (site != 'ANY'):
            options['site'] = site
        
        out_ds = grid_config['outDS']
        if '{HPO_PROJECT_NAME}' in out_ds:
            out_ds = out_ds.format(HPO_PROJECT_NAME=project_name)
        options['outDS'] = out_ds

        command = stylus.join_options(options)
        os.system("phpo {}".format(command))
        
        
        
        
        
        
        
        
        
        
        
        


import sys
import json

try:
    from hpogrid.components.validation import validate_job_metadata
except:
    raise ImportError('Cannot import hpogrid module. Try source setupenv.sh first.')


class MLFlowLogger():

    def __init__(self, exp_name='experiment'):
        from mlflow.tracking import MlflowClient
        self.client = MlflowClient()
        self.experiment_id = self.client.create_experiment(exp_name)
        self.run_ids = []

    def new_run(self):
        run = self.client.create_run(self.experiment_id)
        self.run_ids.append(run.info.run_id)

    def log_hyperparameters(self, hp_dict, run_id=None):
        run_id = run_id or self.run_ids[-1]
        for key, value in hp_dict.items():
            self.client.log_param(run_id, key, value)

    def log_metrics(self, metric_dict, run_id=None):
        run_id = run_id or self.run_ids[-1]
        for key, value in metric_dict.items():
            if not isinstance(value, float):
                continue
            self.client.log_metric(
                run_id, key, value)

    def terminate_run(self, run_id=None):
        run_id = run_id or self.run_ids[-1]
        self.client.set_terminated(run_id)

    def run_server(self):
        from mlflow.cli import cli
        sys.argv = ['mlflow','server']
        sys.exit(cli())

    @classmethod
    def from_json(cls, file):
        with open(file,'r') as log_file:
            log = json.load(log_file)
        if not validate_job_metadata(log):
            raise RuntimeError('Invalid format for hpo result summary.')
        mlflow_logger = cls(log['title'])
        hyperparameters = log['hyperparameters']
        metric = log['metric']
        for i in log['result']:
            hp_dict = {hp:log['result'][i][hp] for hp in hyperparameters}
            metric_dict ={metric:log['result'][i][metric]}
            mlflow_logger.new_run()
            mlflow_logger.log_hyperparameters(hp_dict)
            mlflow_logger.log_metrics(metric_dict)
            mlflow_logger.terminate_run()
        return mlflow_logger



import os
import sys
import json
import time
import importlib
import argparse
from typing import Optional, Dict, List
from datetime import datetime
from pdb import set_trace

import ray
from ray import tune

from hpogrid.components import validation
from hpogrid.components.defaults import *
from hpogrid.utils import helper
    
class JobBuilder():
    def __init__(self):
        self.reset()
        
    @property
    def algorithm(self) -> str:
        return self._algorithm
    
    @algorithm.setter
    def algorithm(self, name):
        if name and (name not in kSearchAlgorithms):
            raise ValueError('unrecognized search algorithm: {}'.format(name))
        self._algorithm = kAlgoMap.get(name, name)
    
    @property
    def scheduler(self) -> str:
        return self._scheduler
    
    @scheduler.setter
    def scheduler(self, name):
        if name and (name not in kSchedulers):
            raise ValueError('unrecognized trial scheduler: {}'.format(name))
        self._scheduler = name
        

    @property
    def mode(self) -> str:
        return self._mode
    
    @mode.setter
    def mode(self, value):
        if value not in kMetricMode:
            raise ValueError('metric mode must be either "min" or "max"')
        self._mode = value

    @property
    def df(self) -> 'pandas.DataFrame':
        return self._df
    
    @property
    def start_datetime(self) -> str:
        return self._start_datetime
    
    @property
    def end_datetime(self) -> str:
        return self._end_datetime
    
    @property
    def start_timestamp(self) -> float:
        return self._start_timestamp
    
    @property
    def total_time(self) -> float:
        return self._total_time
    
    @property
    def best_config(self) -> Dict:
        return self._best_config

    @property
    def hyperparameters(self):
        return self._hyperparameters
    
    
    def reset(self):
        self.project_name = None
        # model configuration
        self.model_script = None
        self.model_name = None
        self.model_param = kDefaultModelParam  
        # search space
        self.search_space = {}
        # hpo configuration
        self._algorithm = None
        self.metric = None
        self._mode = None
        self._scheduler = None  
        self.algorithm_param = kDefaultAlgorithmParam        
        self.scheduler_param = kDefaultSchedulerParam
        self.num_trials = kDefaultTrials
        self.max_concurrent = kDefaultMaxConcurrent
        self.verbose = False
        self.log_dir = kDefaultLogDir
        self.stop = kDefaultStopping
        # parameters for logging
        self._df = None
        self._start_datetime = None
        self._end_datetime = None
        self._total_time = None
        self._best_config = None
        self._hyperparameters = {}
        self._start_timestamp = None
        self.idds_job = False

        
    @staticmethod
    def get_scheduler(name, metric, mode, search_space = None, **args):
        if (name == None) or (name == 'None'):
            return None
        elif name == 'asynchyperband':
            from hpogrid.scheduler.asynchyperband_scheduler import AsyncHyperBandSchedulerWrapper
            return AsyncHyperBandSchedulerWrapper().create(metric, mode, **args)
        elif name == 'bohbhyperband':
            from hpogrid.scheduler.bohbhyperband_scheduler import BOHBHyperBandSchedulerWrapper
            return BOHBHyperBandSchedulerWrapper().create(metric, mode, **args)
        elif name == 'pbt':
            from hpogrid.scheduler.pbt_scheduler import PBTSchedulerWrapper
            if search_space is None:
                raise ValueError('Missing search space definition for pbt scheduler')
            return PBTSchedulerWrapper().create(metric, mode, search_space, **args)

    @staticmethod
    def get_search_space(base_search_space, algorithm):
        if base_search_space is None:
            raise ValueError('search space can not be empty')    
        if algorithm == 'ax':
            from hpogrid.search_space.ax_space import AxSpace
            return AxSpace(base_search_space).get_search_space()
        elif algorithm == 'bohb':
            from hpogrid.search_space.bohb_space import BOHBSpace
            return BOHBSpace(base_search_space).get_search_space()
        elif algorithm == 'hyperopt':
            from hpogrid.search_space.hyperopt_space import HyperOptSpace
            return HyperOptSpace(base_search_space).get_search_space()
        elif algorithm == 'skopt':
            from hpogrid.search_space.skopt_space import SkOptSpace
            return SkOptSpace(base_search_space).get_search_space()
        elif algorithm == 'tune':
            from hpogrid.search_space.tune_space import TuneSpace
            return TuneSpace(base_search_space).get_search_space()
        elif algorithm == 'nevergrad':
            from hpogrid.search_space.nevergrad_space import NeverGradSpace
            return NeverGradSpace(base_search_space).get_search_space() 
        else:
            raise ValueError('Unrecognized search algorithm: {}'.format(name))

    @staticmethod
    def get_algorithm(name, metric, mode, base_search_space, max_concurrent=None, **args):
        if name == 'ax':
            from hpogrid.algorithm.ax_algorithm import AxAlgoWrapper
            algorithm = AxAlgoWrapper().create(metric, mode, base_search_space, **args)
        elif name == 'bohb':
            from hpogrid.algorithm.bohb_algorithm import BOHBAlgoWrapper
            algorithm = BOHBAlgoWrapper().create(metric, mode, base_search_space, **args)
        elif name == 'hyperopt':
            from hpogrid.algorithm.hyperopt_algorithm import HyperOptAlgoWrapper
            algorithm = HyperOptAlgoWrapper().create(metric, mode, base_search_space, **args)
        elif name == 'skopt':
            from hpogrid.algorithm.skopt_algorithm import SkOptAlgoWrapper
            algorithm = SkOptAlgoWrapper().create(metric, mode, base_search_space, **args)
        elif name == 'nevergrad':
            from hpogrid.algorithm.nevergrad_algorithm import NeverGradAlgoWrapper
            algorithm = NeverGradAlgoWrapper().create(metric, mode, base_search_space, **args)
        elif name == 'tune':
            algorithm = None
        else:
            raise ValueError('Unrecognized search algorithm: {}'.format(name))
        # limit max concurrency
        if algorithm and max_concurrent:
            from ray.tune.suggest import ConcurrencyLimiter
            algorithm = ConcurrencyLimiter(algorithm, max_concurrent=max_concurrent) 
        return algorithm

    @staticmethod
    def get_model(script_name, model_name):
        model = None
        script_name_noext = os.path.splitext(script_name)[0]
        try: 
            module = importlib.import_module(script_name_noext)
            model = getattr(module, model_name)
        except: 
            raise ImportError('Unable to import function/class {} '
                'from training script: {}.py'.format(model_name, script_name_noext))
        return model

    def create_metadata(self, df) -> Dict:

        summary = {
            'project_name' : self.project_name,
            'start_datetime': self.start_datetime,
            'end_datetime': self.end_datetime,
            'start_timestamp': self.start_timestamp,
            'task_time_s' : self.total_time,
            'hyperparameters': self.hyperparameters,
            'metric': self.metric,
            'mode' : self.mode,
            'best_config' : self.best_config, 
        }
        
        rename_cols = { 'config/{}'.format(hp): hp for hp in self.hyperparameters}
        rename_cols['time_total_s'] = 'time_s'
        
        df = df.rename(columns=rename_cols)
        cols_to_save = ['time_s'] + self.hyperparameters
        if 'metric' in summary:
            cols_to_save.append(summary['metric'])
        df = df.filter(cols_to_save, axis=1).transpose()
        summary['result'] = df.to_dict()

        return summary
    
    def create_idds_output(self, summary):
        idds_output = {}
        idds_output['status'] = 0
        loss = summary['result'][0][summary['metric']]
        idds_output['loss'] = loss
        idds_output['message'] = ''
        return idds_output

    @staticmethod
    def get_resource_info() -> Dict:
        resource = {}
        n_gpu = helper.get_n_gpu()
        n_cpu = helper.get_n_cpu()
        print('INFO: Number of GPUs detected: ',n_gpu)
        print('INFO: Number of CPUs detected: ',n_cpu)
        if n_gpu > 0:
            resource['gpu'] = n_gpu
            print('INFO: Each trial will use {} GPU(s) resource'.format(n_gpu))
        if n_cpu > 0:
            resource['cpu'] = n_cpu
            print('INFO: Each trial will use {} CPU(s) resource'.format(n_cpu))
        if not resource:
            resource = None
        return resource
    
    def run(self) -> None:
        
        self._start_timestamp = start = time.time()

        if self.algorithm == 'tune':
            tune_config_space = JobBuilder.get_search_space(self.search_space, algorithm='tune')
        else:
            tune_config_space = {}

        tune_config_space.update(self.model_param)

        algorithm = JobBuilder.get_algorithm(
            self.algorithm,
            self.metric,
            self.mode,
            self.search_space,
            self.max_concurrent,
            **self.algorithm_param)
        
    
        scheduler = JobBuilder.get_scheduler(
            self.scheduler,
            self.metric,
            self.mode,
            self.search_space,
            **self.scheduler_param)

        resources_per_trial = JobBuilder.get_resource_info()

        model = JobBuilder.get_model(self.model_script, self.model_name)

        self._start_datetime = datetime.now().strftime("%m/%d/%Y, %H:%M:%S")


        # save something to prevent looping job
        with open(kGridSiteMetadataFileName,'w') as output:
            json.dump({}, output)
        #ray.init(node_ip_address="127.0.0.1", ignore_reinit_error=True)
        try:
            ray.init(ignore_reinit_error=True)

            analysis = tune.run(
                model,
                name=self.project_name,
                scheduler=scheduler,
                search_alg=algorithm,
                config=tune_config_space,
                num_samples=self.num_trials,
                resources_per_trial=resources_per_trial,
                verbose=self.verbose,
                local_dir=self.log_dir,
                stop=self.stop,
                raise_on_failed_trial=False)
        finally:
            ray.shutdown()

        end = time.time()
        self._total_time = float(end-start)
        self._end_datetime = datetime.now().strftime("%m/%d/%Y, %H:%M:%S")

        self._best_config = analysis.get_best_config(metric=self.metric, mode=self.mode)
        
        print("Best config: ", self.best_config)
        print("Time taken in seconds: ", self.total_time)

        df = analysis.dataframe()

        metadata = self.create_metadata(df)
        
        # save metadata
        with open(kGridSiteMetadataFileName,'w') as output:
            json.dump(metadata, output, cls=helper.NpEncoder)
        
        # save idds output
        if self.idds_job:
            idds_output = self.create_idds_output(metadata)
            idds_output_path = os.path.join(kWorkDir,kiDDSHPoutput)
            with open(idds_output_path, 'w') as idds_output_file:
                json.dump(idds_output, idds_output_file, cls=helper.NpEncoder)

        
    @classmethod
    def from_project(cls, proj_name:str, idds_job=False):
        
        config = helper.get_project_config(proj_name)
        
        helper.set_script_path(proj_name)
        
        if idds_job:
            input_point_path = os.path.join(kWorkDir, kiDDSHPinput)
            with open(input_point_path,'r') as input_point_file:
                input_point = json.load(input_point_file)
            return JobBuilder.from_config_for_idds(config, input_point)
        
        return JobBuilder.from_config(config)
    
    
    
    @classmethod
    def from_json(cls, fname:str):
        with open(fname, 'r') as config_file:
            config = json.load(config_file)       
        return JobBuilder.from_config(config)
    
    @classmethod
    def from_config_for_idds(cls, config:Dict, input_point:Dict):
        job = JobBuilder()
        
        job.project_name = config['project_name']
        
        # Load search space
        job.search_space = {}
        
        # Load hpo configuration
        job.algorithm = 'random'
        job.metric = config['hpo_config']['metric']
        job.mode = config['hpo_config']['mode']
        job.scheduler = None
        job.scheduler_param = {}
        job.algorithm_param = {}
        job.num_trials = 1
        job.max_concurrent = 1
        job.verbose = False
        job.log_dir = config['hpo_config']['log_dir']
        job.stop = config['hpo_config']['stop']
        
        # Load model configuration
        job.model_script = config['model_config']['script']
        job.model_name = config['model_config']['model']
        job.model_param = config['model_config']['param']
        
        if set(input_point.keys()) != set(config['search_space'].keys()):
            raise ValueError('hyperparameters given by input search point '
                             'are inconsistent with search space definition')
        
        job.model_param.update(input_point)
        job._hyperparameters = list(input_point.keys())
        job.idds_job = True
    
        return job
       
    
    @classmethod
    def from_config(cls, config:Dict):
        job = JobBuilder()
        
        job.project_name = config['project_name']
        
        # Load search space
        job.search_space = config['search_space']
        
        # Load hpo configuration
        job.algorithm = config['hpo_config']['algorithm']
        job.metric = config['hpo_config']['metric']
        job.mode = config['hpo_config']['mode']
        job.scheduler = config['hpo_config']['scheduler']
        job.scheduler_param = config['hpo_config']['scheduler_param']
        job.algorithm_param = config['hpo_config']['algorithm_param']
        job.num_trials = config['hpo_config']['num_trials']
        job.max_concurrent = config['hpo_config']['max_concurrent']
        job.verbose = config['hpo_config']['verbose']
        job.log_dir = config['hpo_config']['log_dir']
        job.stop = config['hpo_config']['stop']
        
        # Load model configuration
        job.model_script = config['model_config']['script']
        job.model_name = config['model_config']['model']
        job.model_param = config['model_config']['param']

        job._hyperparameters = list(job.search_space.keys())
    
        return job
       
def local_job_parser(idds_job=False):
    parser = argparse.ArgumentParser(description='run hyperparameter optimization '
                                    'for a user defined machine learning project')     
    parser.add_argument('proj_name', help='Name of the project on which '
                        'hyperparameter optimization is run')    
    args = parser.parse_args(sys.argv[2:])

    job = JobBuilder.from_project(args.proj_name, idds_job=idds_job)
    job.run()
            
def grid_job_parser(idds_job=False):
    try:
        extracted_files = helper.extract_tarball(kDataDir, kDataDir)
        helper.grid_site_setup()
        local_job_parser(idds_job=idds_job)
    finally:
        helper.remove_files(extracted_files)
            
    

